/*--------------------------------------------------------------------------
 Demo client for Dr.Web daemon 
                                       ID Antivirus Lab http://www.drweb.ru
                                        http://www.drweb.ru/linux-bsd/forum
                                              http://www.drweb.ru/linux-bsd
--------------------------------------------------------------------------*/

#ifndef lint
static const char rcsid[] = "$Id: drwebdc.c,v 1.17 2001/06/15 03:58:37 cvser Exp $";
#endif

#if defined(__WIN32_DW)
#include <io.h>
#endif /* __WIN32_DW */
#if defined(__UNIX_DW)
#include <sys/param.h>
#include <unistd.h>
#endif /* __UNIX_DW */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include "dwc_lib.h"

#if defined(__WIN32_DW)
#define PATH_MAX	_MAX_PATH
#endif /* __WIN32_DW */

#define DWCLIENT_NOERROR        0     /* 0x00000000 */
#define DWCLIENT_EXPAND         1<<31 /* 0x80000000 */
#define DWCLIENT_DEMO_USAGE     1<<30 /* 0x40000000 */
#define DWCLIENT_NO_MEMORY      1<<29 /* 0x20000000 */
#define DWCLIENT_LIB_ERR        1<<28 /* 0x10000000 */

static char szFile[ PATH_MAX + 2 ] = { 0 };
static char szDaemonIP[ PATH_MAX + 2 ] = { 0 };
static char szSocketFile[ PATH_MAX + 2 ] = { 0 };
static char szPidFile[ PATH_MAX + 2 ] = { 0 };
static int  bShowID = 0;
static int  bShowVersion = 0;
static int  bShowBases = 0;
static int  bScanFile = 0;
static int  bPartFile = 0;
static int  bExpand = 0;
static int  bQuiet = 0;
static int  bDebug = 0;
static int  bNonDiskFile = 0;
static int  scanOptions = 0;
static int  Port = 0;
static int  Timeout = 0;
static size_t Offset = 0;
static size_t Length = 0;

static int  parseCmdLine( int argc, char ** argv );
static int  usage( void );

#if defined(__WIN32_DW)
#define read_file( fd, buf, len )   _read( (fd), (buf), (len) )
#else
#define read_file( fd, buf, len )   read( (fd), (buf), (len) )
#endif

//--------------------------------------------------------------------------
int main( int argc, char *argv[] )
{
    int retcode = DWCLIENT_NOERROR;
    int libcode = DWL_NOERROR;
    char *pFileBuffer = NULL;
    char *pTmp = NULL;

        // parse command line:
    if( !parseCmdLine( argc, argv ) )
    {
        return usage();
    }
    outputLib( bQuiet, bDebug, 0 );
    if( (szDaemonIP[0] != 0) || (Port > 0) )
    {
        libcode = startupLib( DWL_TYPE_TCP_SOCKET, Timeout, szDaemonIP, Port );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("Can't startup library with tcp/ip transport.\n");
            return DWCLIENT_LIB_ERR;
        }
    }
#if defined(__UNIX_DW)
    else if( szSocketFile[0] != 0 )
    {
        libcode = startupLib( DWL_TYPE_UNIX_SOCKET, Timeout, szSocketFile );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("Can't startup library with unix sockets transport.\n");
            return DWCLIENT_LIB_ERR;
        }
    }
    else if( szPidFile[0] != 0 )
    {
        libcode = startupLib( DWL_TYPE_READPIDFILE, Timeout, szPidFile );
        if( libcode != DWL_NOERROR )
        {
            wprintLib("Can't startup library by reading pidfile.\n");
            return DWCLIENT_LIB_ERR;
        }        
    }
#endif /* __UNIX_DW */
    else
    {
        dprintLib("Daemon address not defined. Try use all by defaults.\n");
        libcode = startupLib( DWL_TYPE_DEFAULT, Timeout );
        if( libcode != DWL_NOERROR )         
        {
            wprintLib("Can't startup library by.\n");
            return DWCLIENT_LIB_ERR;
        }        
    }
    if( bShowVersion )
    {
        libcode = showVersion( NULL, NULL );
        if( libcode != DWL_NOERROR )
        {
            retcode = DWCLIENT_LIB_ERR;
            goto error_exit;
        }
    }
    if( bShowID )
    {
        libcode = showID( NULL, NULL );
        if( libcode != DWL_NOERROR )
        {
            retcode = DWCLIENT_LIB_ERR;
            goto error_exit;
        }
    }
    if( bShowBases )
    {
        libcode = showBases( NULL, NULL );
        if( libcode != DWL_NOERROR )
        {
            retcode = DWCLIENT_LIB_ERR;
            goto error_exit;
        }
    }
    if( bScanFile || bPartFile )
    {
        if( szFile[0] != 0 )
        {
#if defined(__UNIX_DW)
            if( bExpand )
            {
                char buf[ PATH_MAX + 2 ];
                if( realpath( szFile, buf ) == NULL )
                {
                    wprintLib("Can't expand name.\n");
                    retcode = DWCLIENT_EXPAND;
                    goto error_exit;
                }
                strncpy( szFile, buf, PATH_MAX + 1 );
            }

            if( (szFile[0] != '/' && !bPartFile) || bNonDiskFile )
            {
#endif /* __UNIX_DW */
                libcode = scanFile( szFile, strlen(szFile), DWL_FT_NORMALFILE, scanOptions, &retcode, NULL, NULL );
                if( libcode != DWL_NOERROR )
                {
                    retcode = DWCLIENT_LIB_ERR;
                }
#if defined(__UNIX_DW)
    	    }
    	    else
    	    {	
                if( bPartFile )
                {
                    libcode = scanFilePart( szFile, strlen(szFile), Offset, Length, scanOptions, &retcode, NULL, NULL );
                    if( libcode != DWL_NOERROR )
                    {
                        retcode = DWCLIENT_LIB_ERR;
                    }
                }
                else                
                {
                    libcode = scanFile( szFile, strlen(szFile), DWL_FT_DISKFILE, scanOptions, &retcode, NULL, NULL );
                    if( libcode != DWL_NOERROR )
                    {
                        retcode = DWCLIENT_LIB_ERR;
                    }
                }
    	    }
#endif /* __UNIX_DW */
        }
        else
        {
            char buf[ 0x4000 ];
            int fileSize = 0;
            int readed = read_file( fileno(stdin), buf, sizeof(buf) );
            while( readed > 0 )
            {
                pTmp = realloc( pFileBuffer, fileSize + readed );
                if( !pTmp )
                {
                    wprintLib("No memory to expand for readed file.\n");
                    free( pFileBuffer );
                    retcode = DWCLIENT_NO_MEMORY;
                    goto error_exit;
                }
                memcpy( pTmp + fileSize, buf, readed );
                pFileBuffer = pTmp;
                fileSize += readed;
                readed = read_file( fileno(stdin), buf, sizeof(buf) );
            }
            libcode = scanBuffer( pFileBuffer, fileSize, scanOptions, &retcode, NULL, NULL );
            if( libcode != DWL_NOERROR )
            {
                retcode = DWCLIENT_LIB_ERR;
            }
        }
    }

  error_exit:
    cleanupLib();
    if( pFileBuffer )
    {
        free( pFileBuffer );
        pFileBuffer = NULL;
    }
    if( (retcode & ~DERR_MASK ) != 0 ) /* have local errors */
    {
        if( retcode & DWCLIENT_LIB_ERR )
        {
            dprintLib("Results: dwclient library error %d (%s)\n", libcode, getLibError( libcode ) );
        }
    }
    else
    {
        if( retcode != DERR_NOERROR ) /* scan not ok - errors or viruses */
        {
            dprintLib("Results: daemon return code %#x (%s)\n", retcode, getDaemonCode( retcode ) );
        }
        else
        {
            dprintLib("Results: daemon return ok.\n");
        }
    }
    dprintLib("Return code %x\n", retcode );
    return retcode;
}

//--------------------------------------------------------------------------
static int parseCmdLine( int argc, char ** argv )
{
#if defined(__UNIX_DW)

#define DWC_PCO_NAME    (1<<0)
#define DWC_PCO_OFF     (1<<1)
#define DWC_PCO_LEN     (1<<2)
#define DWC_PCO_FULL    (DWC_PCO_NAME | DWC_PCO_OFF | DWC_PCO_LEN)
    int  PartCmdOpt = 0, intVal, Cont;

    Cont = 1;
    while( Cont )
    {
        switch( getopt( argc, argv, "exihqws:r:t:n:p:d:u:c:o:l:f:") )
        {
            case 'f':
                if( bScanFile )
                {
                    wprintLib("You can not define -f and -i option together.\n");
                    return 0;
                } 
                if( optarg )
                {
                    strncpy( szFile, optarg, sizeof(szFile)-1 );
                }
                bScanFile = 1;
                break;
            case 'i':
                if( bScanFile )
                {
                    wprintLib("You can not define -f and -i option together.\n");
                    return 0;
                } 
                bScanFile = 1;
                break;
            case 'd':
                strncpy( szPidFile, optarg, sizeof(szPidFile)-1 );
                break;
            case 'u':
                strncpy( szSocketFile, optarg, sizeof(szSocketFile)-1 );
                break;
            case 'c':
                strncpy( szFile, optarg, sizeof(szFile)-1 );
                PartCmdOpt |= DWC_PCO_NAME;
                bScanFile = 1;
                bPartFile = 1;
                break;
            case 'o':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    wprintLib("Error read offset argument (-o option).\n");
                    return 0;
                }
                intVal = atoi( optarg );         
                if( intVal >= 0 )
                {
                    Offset = intVal;
                    PartCmdOpt |= DWC_PCO_OFF;
                    bExpand = 1;
                }
                break;
            case 'l':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    wprintLib("Error read length argument (-l option).\n");
                    return 0;
                }
                intVal = atoi( optarg );         
                if( intVal >= 0 )
                {
                    Length = intVal;
                    PartCmdOpt |= DWC_PCO_LEN;
                }
                break;
            case 'n':
                strncpy( szDaemonIP, optarg, sizeof(szDaemonIP)-1 );
                break;
            case 'p':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    wprintLib("Error read port argument (-p option).\n");
                    return 0;
                }
                intVal = atoi( optarg );         
                if( intVal >= 0 )
                {
                    Port = intVal;
                }
                break;
            case 't':
                intVal = optarg[0];
                if( !isdigit( intVal ) )
                {
                    wprintLib("Error read timeout argument (-t timeout).\n");
                    return 0;
                }
                intVal = atoi( optarg );         
                if( intVal >= 0 )
                {
                    Timeout = intVal * 1000;
                }
                break;
            case 'h':
                scanOptions |= DRWEBD_HEURISTIC_ON;
                break;
            case 'e':
                bExpand = 1;
                break;
            case 'x':
                bNonDiskFile = 1;
                break;
            case 's':
                switch( optarg[0] )
                {
                    case 'v':
                        bShowVersion = 1;
                        break;
                    case 'i':
                        bShowID = 1;
                        break;
                    case 'b':
                        bShowBases = 1;
                        break;
                    default: 
                        wprintLib("\nInvalid option show (-s?)");     
                        return 0;           
                }
                break;
            case 'r':
                switch( optarg[0] )
                {
                    case 'v':
                        scanOptions |= DRWEBD_RETURN_VIRUSES;
                        break;
                    case 'c':
                        scanOptions |= DRWEBD_RETURN_CODES;
                        break;
                    case 'r':
                        scanOptions |= DRWEBD_RETURN_REPORT;
                        break;
                    default: 
                        wprintLib("\nInvalid option return (-r?)");                
                        return 0;           
                }
                break;
            case 'q':
                bQuiet = 1;
                break;
            case 'w':
                bDebug = 1;
                break;
            case -1 :
                if( bPartFile && ((PartCmdOpt & DWC_PCO_FULL) != DWC_PCO_FULL) )
                {
                    wprintLib( "\nFor scan only part file use SHOULD define -c -o -l options." );
                    return 0;
                }
                Cont = 0;
                break;
            default :
                return 0;
        }
    }    
#else /* !__UNIX_DW */
    int i;

    for( i = 1; i < argc; i++ )
    {
        if( !strncmp( *(argv + i), "-f", 2 ) )
        {
            if( strlen( *(argv + i) ) > 2 )
            {
                strncpy( szFile, *( argv + i ) + 2, PATH_MAX );
            }
            bScanFile = 1;
        }
        else if( !strncmp( *(argv + i), "-n", 2 )
                 &&
                 strlen( *(argv + i) ) > 2 )
        {
            strncpy( szDaemonIP, *( argv + i ) + 2, PATH_MAX );
        }
        else if( !strncmp( *(argv + i), "-p", 2 )
                 &&
                 strlen( *(argv + i) ) > 2 )
        {
            Port = atoi( *( argv + i ) + 2 );
        }
        else if( !strncmp( *(argv + i), "-t", 2 )
                 &&
                 strlen( *(argv + i) ) > 2 )
        {
            Timeout = 1000 * atoi( *( argv + i ) + 2 );
        }
        else if( !strcmp( *(argv + i), "-h" ) )
        {
            scanOptions |= DRWEBD_HEURISTIC_ON;
        }
        else if( !strcmp( *(argv + i), "-rv" ) )
        {
            scanOptions |= DRWEBD_RETURN_VIRUSES;
        }
        else if( !strcmp( *(argv + i), "-rr" ) )
        {
            scanOptions |= DRWEBD_RETURN_REPORT;
        }
        else if( !strcmp( *(argv + i), "-rc" ) )
        {
            scanOptions |= DRWEBD_RETURN_CODES;
        }
        else if( !strcmp( *(argv + i), "-q" ) )
        {
            bQuiet = 1;
        }
        else if( !strcmp( *(argv + i), "-w" ) )
        {
            bDebug = 1;
        }
        else if( !strcmp( *(argv + i), "-sb" ) )
        {
            bShowBases = 1;
        }
        else if( !strcmp( *(argv + i), "-si" ) )
        {
            bShowID = 1;
        }
        else if( !strcmp( *(argv + i), "-sv" ) )
        {
            bShowVersion = 1;
        }
        else
        {
            return 0;
        }
    }
#endif /* __UNIX_DW */    
    return argc > 1;
}

//--------------------------------------------------------------------------
static int usage( void )
{
    printf( "\nUsage: drwebdc [options]"
            "\n"
#if defined(__WIN32_DW)            
            "\n -f[FILE_NAME]    file to scan (STDIN is default)"
#else /* __WIN32_DW */
            "\n -fFILE_NAME      file to scan"
            "\n -i               read from stdin"
#endif /* __WIN32_DW) */           
#if defined(__UNIX_DW)
            "\n -dPIDFILE_NAME   Dr.Web daemon PID-file (/var/run/drwebd.pid is default)"
            "\n -uSOCKET_FILE    Dr.Web daemon UNIX-socket file"
            "\n -cFILE_NAME      name file to part-scan (must to define offset and length)"
            "\n -oOFFSET         offset for file to part-scan"
            "\n -lLENGTH         length for file to part-scan"
#endif /* __UNIX_DW */    
            "\n -nINETADDRESS    Dr.Web daemon Internet address (IP or domain name)"
            "\n -pPORT           Dr.Web daemon TCP port"
            "\n -tTIMEOUT        timeout (in millisec) for communcations operations"
            "\n -h               use heuristic analyser"
#if defined(__UNIX_DW)
            "\n -e               expand file name (automatic with -o option)"
            "\n -x               dont scan as local file, even if it given with full path"
#endif /* __UNIX_DW */    
            "\n -sb              show virus bases info"
            "\n -si              show daemon id string"
            "\n -sv              show daemon version"
            "\n -rv              return virus names"
            "\n -rc              return extended codes"
            "\n -rr              return report lines"
            "\n -q               quiet mode"
            "\n -w               debug mode"
            "\n" );
    return DWCLIENT_DEMO_USAGE;
}
