#ifndef _DWC_LOG_C
#define _DWC_LOG_C

#ifndef lint
static const char rcsid[] = "$Id: dwc_log.c,v 1.4 2001/06/14 06:06:34 cvser Exp $";
#endif

#include <stdio.h>
#include <stdarg.h>
#include "dwc_log.h"

DWC_bool_t VideoLogEnabled;
DWC_bool_t DebugLogEnabled;
DWC_bool_t LogSyslog;

//--------------------------------------------------------------------------
DWC_string_t findWhat( int what, DWC_msg_entry_t Table[] )
{
    DWC_msg_entry_t *iterator;
   
    for( iterator = Table; iterator->name; ++iterator )
        if( iterator->code == what )
            return iterator->name;
    return "unknown code";
}

/* -------------------------------------------------- */
void InternalPrint( int facility, const char* fmt, va_list args )
{
    if( !LogSyslog )
    {
        vfprintf( stdout, fmt, args );
    }
#if defined(__UNIX_DW)
    else
    {
        vsyslog( LOG_MAIL | facility, fmt, args );
    }
#endif /* __UNIX_DW */
}

/* -------------------------------------------------- */
void StartLog( DWC_log_t* log )
{
    if( log && (log->max_len >= 0) )
    {
        log->line[0] = NULLCHAR;
        log->cur_len = 0;
    }
}

/* -------------------------------------------------- */
void AddLog( DWC_log_t* log, DWC_string_t fmt, ... )
{
    va_list args;
    int     add;

    va_start( args, fmt );
    if( log )
    {
        if( (0 <= log->cur_len) && (log->cur_len < log->max_len) )
        {
#if defined(__WIN32_DW)
            add = _vsnprintf( &log->line[log->cur_len], log->max_len - log->cur_len, fmt, args );
#else  /* __WIN32_DW */
            add = vsnprintf( &log->line[log->cur_len], log->max_len - log->cur_len, fmt, args );
#endif /* __WIN32_DW */
            if( add >= 0 )
            {
                log->cur_len += add;
            }
            else
            {
                log->line[log->cur_len] = NULLCHAR;
            }
        }
    }
    else
    {
        InternalPrint( LOG_NOTICE, fmt, args );
    }
    va_end( args );
}

/* -------------------------------------------------- */
void EndLog( DWC_log_t* log )
{
    if( log && (log->max_len > 0) )
    {
        if( (0 < log->cur_len) && (log->cur_len < log->max_len)  )
        {
            log->line[log->cur_len] = NULLCHAR;
        }
        else
        {
            log->line[log->max_len - 1] = NULLCHAR;
        }
    }
}

#endif /* _DWC_LOG_H */
