#ifndef __DWC_TCP_C
#define __DWC_TCP_C

#ifndef lint
static const char rcsid[] = "$Id: dwc_tcp.c,v 1.12 2001/06/19 00:11:31 cvser Exp $";
#endif

#include <string.h>
#include <errno.h>

#if defined(__UNIX_DW)
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#if defined(__THREAD_BASED_DW)
#include <pthread.h>
#endif /* __THREAD_BASED_DW */
#endif /* __UNIX_DW */

#include "dwc_err.h"
#include "dwc_tcp.h"
#include "dwc_macro.h"
#include "dwc_log.h"

#if defined(__THREAD_BASED_DW)
#if defined(__UNIX_DW)
pthread_mutex_t  resolve_mutex;
#endif /* __UNIX_DW */
#endif /* __THREAD_BASED_DW */

static int tcp_error;

/*
 * Description:
 *  String Table with Windows Socket errors as integer identifiers and
 *  Error value and text description in text (string resources).  With
 *  this table as part of a .RC file, the LoadString() function can be
 *  used to retrieve the strings by referencing the error number.
 * Edit History:
 *  18-Feb-93  <MIKE@sx.gar.no> created and sent to WinSockAPI mailing list 
 *				 for public distribution.
 *  25-Apr-93  <rcq@ftp.com>    WSAINITIALIZED respelt as WSAINITIALISED,
 *				 and added header with comments.
 */
#if defined(__WIN32_DW) 
static DWC_msg_entry_t  WinsockErrMsgTable[] = {
  { WSABASEERR,         "[0] No Error" },
  { WSAEINTR,           "[10004] Interrupted system call" },
  { WSAEBADF,           "[10009] Bad file number" },
  { WSAEACCES,          "[10013] Permission denied" },
  { WSAEFAULT,          "[10014] Bad address" },
  { WSAEINVAL,          "[10022] Invalid argument" },
  { WSAEMFILE,          "[10024] Too many open files" },
  { WSAEWOULDBLOCK,     "[10035] Operation would block" },
  { WSAEINPROGRESS,     "[10036] Operation now in progress" },
  { WSAEALREADY,        "[10037] Operation already in progress" },
  { WSAENOTSOCK,        "[10038] Socket operation on non-socket" },
  { WSAEDESTADDRREQ,    "[10039] Destination address required" },
  { WSAEMSGSIZE,        "[10040] Message too long" },
  { WSAEPROTOTYPE,      "[10041] Protocol wrong type for socket" },
  { WSAENOPROTOOPT,     "[10042] Bad protocol option" },
  { WSAEPROTONOSUPPORT, "[10043] Protocol not supported" },
  { WSAESOCKTNOSUPPORT, "[10044] Socket type not supported" },
  { WSAEOPNOTSUPP,      "[10045] Operation not supported on socket" },
  { WSAEPFNOSUPPORT,    "[10046] Protocol family not supported" },
  { WSAEAFNOSUPPORT,    "[10047] Address family not supported by protocol family" },
  { WSAEADDRINUSE,      "[10048] Address already in use" },
  { WSAEADDRNOTAVAIL,   "[10049] Can't assign requested address" },
  { WSAENETDOWN,        "[10050] Network is down" },
  { WSAENETUNREACH,     "[10051] Network is unreachable" },
  { WSAENETRESET,       "[10052] Net dropped connection or reset" },
  { WSAECONNABORTED,    "[10053] Software caused connection abort" },
  { WSAECONNRESET,      "[10054] Connection reset by peer" },
  { WSAENOBUFS,         "[10055] No buffer space available" },
  { WSAEISCONN,         "[10056] Socket is already connected" },
  { WSAENOTCONN,        "[10057] Socket is not connected" },
  { WSAESHUTDOWN,       "[10058] Can't send after socket shutdown" },
  { WSAETOOMANYREFS,    "[10059] Too many references, can't splice" },
  { WSAETIMEDOUT,       "[10060] Connection timed out" },
  { WSAECONNREFUSED,    "[10061] Connection refused" },
  { WSAELOOP,           "[10062] Too many levels of symbolic links" },
  { WSAENAMETOOLONG,    "[10063] File name too long" },
  { WSAEHOSTDOWN,       "[10064] Host is down" },
  { WSAEHOSTUNREACH,    "[10065] No Route to Host" },
  { WSAENOTEMPTY,       "[10066] Directory not empty" },
  { WSAEPROCLIM,        "[10067] Too many processes" },
  { WSAEUSERS,          "[10068] Too many users" },
  { WSAEDQUOT,          "[10069] Disc Quota Exceeded" },
  { WSAESTALE,          "[10070] Stale NFS file handle" },
  { WSAEREMOTE,         "[10071] Too many levels of remote in path" },
  { WSASYSNOTREADY,     "[10091] Network SubSystem is unavailable" },
  { WSAVERNOTSUPPORTED, "[10092] WINSOCK DLL Version out of range" },
  { WSANOTINITIALISED,  "[10093] Successful WSASTARTUP not yet performed" },
  { WSAHOST_NOT_FOUND,  "[11001] Host not found" },
  { WSATRY_AGAIN,       "[11002] Non-Authoritative Host not found" },
  { WSANO_RECOVERY,     "[11003] Non-Recoverable errors: FORMERR, REFUSED, NOTIMP" },
  { WSANO_DATA,         "[11004] Valid name, no data record of requested" },
  END_OF_TABLE
};
static const int  WinsockErrMsgTableSize = (sizeof(WinsockErrMsgTable) / sizeof(DWC_msg_entry_t)) - 1;
#endif /* __WIN32_DW */

/* -------------------------------------------------------------------- */
DWC_error_code_t init_tcp_sockets ( void )
{
#if defined(__UNIX_DW)
#if defined(__THREAD_BASED_DW)
    if( pthread_mutex_init( &resolve_mutex, NULL ) != 0 )
    {
        dprintLib( "TCP sockets: cant init mutex.\n" );
        return DWC_NORESOURCE;
    }
#endif /* __THREAD_BASED_DW */    
    return DWC_NOERROR;
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)

    WORD    wVersionRequested;
    WSADATA wsaData;
   
    wVersionRequested = MAKEWORD( 1, 1 );
    if( WSAStartup( wVersionRequested, &wsaData ) != 0 ) 
    {
        dprintLib("TCP sockets: can't init WinSock subsystem.\n");
        SET_SOCKET_ERROR( tcp_error );
        return DWC_IPCSYSTEM;
    }
    if( (LOBYTE( wsaData.wVersion ) != 1)
        ||
        (HIBYTE( wsaData.wVersion ) != 1) ) 
    {
        dprintLib("TCP sockets: Couldn't find usable WinSock (version 1.1) DLL.\n");
        tcp_error = WSACleanup();
        return DWC_IPCSYSTEM; 
    }
    return DWC_NOERROR;
#endif /* __WIN32_DW */
}

/* -------------------------------------------------------------------- */
DWC_error_code_t dinit_tcp_sockets ( void )
{
#if defined(__UNIX_DW)
#if defined(__THREAD_BASED_DW)
    if( pthread_mutex_unlock( &resolve_mutex ) != 0 )
    {
        dprintLib( "TCP sockets: cant unlock mutex.\n" );
    }
    if( pthread_mutex_destroy( &resolve_mutex ) != 0 )
    {
        dprintLib( "TCP sockets: cant destroy mutex.\n" );
        return DWC_IPCSYSTEM;
    }
#endif /* __THREAD_BASED_DW */    
    return DWC_NOERROR;
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
    tcp_error = WSACleanup();
	if( tcp_error != 0 )
	{
	    dprintLib("TCP sockets: can't deinit WinSock subsystem.\n");
        return DWC_IPCSYSTEM;
	}
#endif /* __WIN32_DW */
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_string_t last_tcp_error ( void )
{
#if defined(__WIN32_DW)
    static char buf[128];
    int         err;

    err = WSAGetLastError();
    _snprintf( buf, sizeof(buf)-1, "WinSockError(%d): %s", err, findWhat(err, WinsockErrMsgTable) );
    return buf;
#else /* !__WIN32_DW */
    return strerror( tcp_error );
#endif /* __WIN32_DW */
}

/* -------------------------------------------------------------------- */
DWC_error_code_t addr_tcp_op ( DWC_cmd_t cmd, DWC_tcp_socket_addr_t* addr, DWC_string_t name, va_list args )
{
#if defined(__UNIX_DW) && defined(__THREAD_BASED_DW)
    int                  retcode;
#endif /* __UNIX_DW && __THREAD_BASED_DW */
    unsigned short       port;
    struct hostent      *he = NULL;
    DWC_bool_t           used = FALSE;
    DWC_error_code_t     error;
    
    if( (cmd == DWC_IPC_CMD_SETADDR)
        ||
        (cmd == DWC_IPC_CMD_SETALL) )
    {
        used = TRUE;
        addr->sin_addr = va_arg( args, struct in_addr );
        if( addr->sin_addr.s_addr == INADDR_NONE )
        {
            error = DWC_BADPARAMS;
            goto error_exit;
        }
    }
    if( (cmd == DWC_IPC_CMD_SETPORT)
        ||
        (cmd == DWC_IPC_CMD_SETALL) )
    {
        used = TRUE;
        port = (unsigned short)(0xFFFF & va_arg( args, unsigned int ));
        if( port == 0 )
        {
            error = DWC_BADPARAMS;
            goto error_exit;
        }
        addr->sin_port = htons( port );
    }
    if( cmd == DWC_IPC_CMD_RESOLVE )
    {
        used = TRUE;
		if( !name )
		{
			tcp_error = EINVAL;
			error = DWC_BADPARAMS;
			goto error_exit;
		}
#if defined(__WIN32_DW)        
        addr->sin_addr.s_addr = inet_addr( name );        
        if( addr->sin_addr.s_addr == INADDR_NONE )
        {
#else /* !__WIN32_DW */
        if( inet_aton( name, &addr->sin_addr ) == 0 )
        {
#endif /* __WIN32_DW */
#if defined(__THREAD_BASED_DW)
#if defined(__UNIX_DW)
            retcode = pthread_mutex_lock( &resolve_mutex );
            if( retcode != 0 )
            {
                dprintLib("tcp_get_host: cant lock mutex on resolver.\n");
                tcp_error = retcode;
                error = DWC_IPCSYSTEM;
                goto error_exit;
            }
#endif /* __UNIX_DW */
#endif /* __THREAD_BASED_DW */    
            he = gethostbyname( name );                        
            if( !he || (he->h_addrtype != AF_INET) )
            {
                dprintLib("tcp_get_host: error(%d) on gethostbyname().\n", h_errno );
                tcp_error = h_errno;
                error = DWC_INVALIDADDR;
                goto error_exit;
            }
            memcpy( &addr->sin_addr, he->h_addr_list[0], DWC_Min(he->h_length, sizeof(addr->sin_addr)) );            
#if defined(__UNIX_DW)
            endhostent();
#endif /* __UNIX_DW */            
#if defined(__THREAD_BASED_DW)
#if defined(__UNIX_DW) 
            retcode = pthread_mutex_unlock( &resolve_mutex );
            if( retcode != 0 )
            {
                dprintLib("tcp_get_host: cant unlock mutex on resolver.\n");
                tcp_error = retcode;
                error = DWC_IPCSYSTEM;
                goto error_exit;
            }
#endif /* __UNIX_DW */
#endif /* __THREAD_BASED_DW */    
        }
    }
    if( !used )
    {
        error = DWC_BADPARAMS;
        goto error_exit;
    }
    error = DWC_NOERROR;
  error_exit:
    return error;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t create_tcp_socket ( DWC_tcp_socket_t *sock, DWC_bool_t nonblocking )
{
#if defined(__WIN32_DW)
    int         err;
#endif /* __WIN32_DW */
#if defined(__UNIX_DW)
    int         opt;
#endif /* __UNIX_DW */

    PARAM_ASSERT( sock );
    
    *sock = socket( PF_INET, SOCK_STREAM, IPPROTO_TCP );
    if( (*sock) == INVALID_TCP_SOCKET )
    {
        dprintLib("Socket() on TCP socket.\n");
        SET_SOCKET_ERROR( tcp_error );
        return DWC_IPCSYSTEM;
    }
    if( nonblocking )
    {
#if defined(__UNIX_DW)
        opt = fcntl( (*sock), F_GETFL, 0 );
        if( fcntl( (*sock), F_SETFL, opt | O_NONBLOCK ) == -1 )
        {
#endif /* __UNIX_DW */
#if defined(__WIN32_DW)
        err = ioctlsocket( (*sock), FIONBIO, &nonblocking );
        if( SOCK_ERR( err ) )
        {
#endif /* __WIN32_DW */
            dprintLib("Can't set nonblocking mode for TCP socket.\n");
            SET_SOCKET_ERROR( tcp_error );
            close_tcp_socket( sock );
            return DWC_IPCSYSTEM;
        }
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t bind_tcp_socket ( DWC_tcp_socket_t sock, DWC_tcp_socket_addr_t* addr )
{
    DWC_socklen_t len = sizeof(DWC_tcp_socket_addr_t);
    
    if( bind( sock, (struct sockaddr*) addr, len ) < 0 )
    {
        dprintLib("Can't bind socket.\n");
        SET_SOCKET_ERROR( tcp_error );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t listen_tcp_socket ( DWC_tcp_socket_t sock, int backlog )
{
    if( listen( sock, backlog ) != 0 )
    {
        dprintLib("Can't listen TCP socket.\n");
        SET_SOCKET_ERROR( tcp_error );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t accept_tcp_socket ( DWC_tcp_socket_t sock, DWC_tcp_socket_addr_t *addr, DWC_tcp_socket_t *newsock )
{
    DWC_socklen_t len = sizeof(DWC_tcp_socket_addr_t);
    
    *newsock = accept( sock, (struct sockaddr*) addr, &len );
    if( (*newsock) == INVALID_TCP_SOCKET )
    {
        dprintLib("Can't accept TCP socket.\n");
        SET_SOCKET_ERROR( tcp_error );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t connect_tcp_socket ( DWC_tcp_socket_t sock, DWC_tcp_socket_addr_t *addr, DWC_timeout_t timeout )
{
    DWC_error_code_t  err;
    int               how;

    if( connect( sock, (struct sockaddr*) addr, sizeof(DWC_tcp_socket_addr_t) ) < 0 )
    {
        if( IN_PROGRESS )
        {
            how = DWC_IPC_WAIT_EXCP | DWC_IPC_WAIT_SEND;
            err = wait_tcp_socket( sock, &how, timeout );
            if( err == DWC_NOERROR  )
            {
                if( how & DWC_IPC_WAIT_SEND )
                {
                    return DWC_NOERROR;
                }
                dprintLib("Connection failed.\n");
                err = DWC_IPCSYSTEM;
            }
#if defined(__WIN32_DW)
            WSASetLastError(WSAECONNREFUSED);
#endif /* __WIN32_DW */
            SET_SOCKET_ERROR( tcp_error );
            return err;
        }
        dprintLib("Can't connect() on TCP socket.\n");
        SET_SOCKET_ERROR( tcp_error );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t send_tcp_socket ( DWC_tcp_socket_t sock, DWC_pointer_t buf, DWC_length_t *len, int options )
{
    int sended;
    
    sended = send( sock, buf, *len, 0 );
    if( SOCK_ERR(sended) )
    {
        (*len) = 0;
        if( WOULD_BLOCK )
        {
            return DWC_NOERROR;
        }
        dprintLib("Can't send() on TCP socket.\n");
        SET_SOCKET_ERROR( tcp_error );
        return DWC_IPCSYSTEM;
    }
    (*len) = sended;
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t recv_tcp_socket ( DWC_tcp_socket_t sock, DWC_pointer_t buf, DWC_length_t *len, int options )
{
    int readed;
     
    readed = recv( sock, buf, *len, options );
    if( SOCK_ERR(readed) )
    {
        (*len) = 0;
        if( WOULD_BLOCK )
        {
            return DWC_NOERROR;
        }
        dprintLib("Can't recv() on TCP socket.\n");
        SET_SOCKET_ERROR( tcp_error );
        return DWC_IPCSYSTEM;
    }    
    (*len) = readed;
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t wait_tcp_socket( DWC_tcp_socket_t sock, int *how, DWC_timeout_t timeout )
{
    fd_set         rd_set, wr_set, ex_set;    
    struct timeval timo, *pt;
    int            rd;
    
    FD_ZERO( &rd_set );
    FD_ZERO( &wr_set );
    FD_ZERO( &ex_set );
    if( (*how & DWC_IPC_WAIT_RECV) == DWC_IPC_WAIT_RECV )
    {
        FD_SET( sock, &rd_set );
    }
    if( (*how & DWC_IPC_WAIT_SEND) == DWC_IPC_WAIT_SEND )
    {
        FD_SET( sock, &wr_set );
    }
    if( (*how & DWC_IPC_WAIT_EXCP) == DWC_IPC_WAIT_EXCP )
    {
        FD_SET( sock, &ex_set );
    }
    *how = 0;
    pt = NULL;
    if( timeout > 0 )
    {
        timo.tv_sec = timeout / 1000000;
        timo.tv_usec = timeout % 1000000;
        pt = &timo;
    }
    rd = select( sock+1, &rd_set, &wr_set, &ex_set, pt );
    if( (rd == 0) && pt )
    {
#if defined(__WIN32_DW)
        WSASetLastError(WSAETIMEDOUT);
#endif /* __WIN32_DW */        
        return DWC_TIMED;
    }
    if( rd < 0 )
    {
        dprintLib("Error select() on TCP socket.\n");
        SET_SOCKET_ERROR( tcp_error );
        return DWC_IPCSYSTEM;
    }
    if( FD_ISSET( sock, &rd_set ) )
    {
        *how |= DWC_IPC_WAIT_RECV;
    }
    if( FD_ISSET( sock, &wr_set ) )
    {
        *how |= DWC_IPC_WAIT_SEND;
    }
    if( FD_ISSET( sock, &ex_set ) )
    {
        *how |= DWC_IPC_WAIT_EXCP;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t shutdown_tcp_socket ( DWC_tcp_socket_t *sock )
{
    int error;

#if defined(__WIN32_DW)
#ifdef SD_BOTH
#define SHUT_RDWR        SD_BOTH
#else /* SD_BOTH */
#define SHUT_RDWR        2
#endif /* SD_BOTH */
#endif /* __WIN32_DW */

    error = shutdown( *sock, SHUT_RDWR );
    (*sock) = INVALID_TCP_SOCKET;
    if( error )
    {
        dprintLib("Can't shutdown() on TCP socket.\n");
        SET_SOCKET_ERROR( tcp_error );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------------------------- */
DWC_error_code_t close_tcp_socket ( DWC_tcp_socket_t *sock )
{
    int error;
    
#if defined(__WIN32_DW)
    error = closesocket( *sock );
#else
    error = close( *sock );
#endif
    *sock = INVALID_TCP_SOCKET;
    if( error )
    {
        dprintLib("Can't close() on TCP socket.\n");
        SET_SOCKET_ERROR( tcp_error );
        return DWC_IPCSYSTEM;
    }
    return DWC_NOERROR;
}

/* -------------------------------------------------- */
void cancel_tcp_call( DWC_pointer_t sock )
{
#if defined(__WIN32_DW)
#ifdef SD_BOTH
#define SHUT_RDWR        SD_BOTH
#else /* SD_BOTH */
#define SHUT_RDWR        2
#endif /* SD_BOTH */
#else /* !__WIN32_DW */
#ifndef SHUT_RDWR
#define SHUT_RDWR       2
#endif /* SHUT_RDWR */
#endif /* __WIN32_DW */
    int  err;

    err = shutdown( *((DWC_tcp_socket_t*)sock), SHUT_RDWR ); 
    if( SOCK_ERR( err ) )
    {
        dprintLib("Can't shutdown (cancel blocking syscall) on TCP socket.\n");
        SET_SOCKET_ERROR( tcp_error );
    }
}

#endif /* __DWC_TCP_H */
