#ifndef __DWC_TCP_H
#define __DWC_TCP_H

#include <stdarg.h>
#include "dwc_type.h"
#include "dwc_timo.h"

#if defined(__WIN32_DW) 

#include <winsock.h>

#define INVALID_TCP_SOCKET      INVALID_SOCKET
#define WOULD_BLOCK             (WSAGetLastError() == WSAEWOULDBLOCK)
#define IN_PROGRESS             (WOULD_BLOCK)
#define SOCK_ERR( retval )      ((retval) == SOCKET_ERROR)
#define SET_SOCKET_ERROR( err ) ((err) = WSAGetLastError())

typedef SOCKET                  DWC_tcp_socket_t;

#else /* not __WIN32_DW */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define INVALID_TCP_SOCKET      (-1)
#define WOULD_BLOCK             (errno == EWOULDBLOCK)
#define IN_PROGRESS             (errno == EINPROGRESS)
#define SOCK_ERR( retval )      ((retval) <= 0)
#define SET_SOCKET_ERROR( err ) ((err) = errno)

typedef int                     DWC_tcp_socket_t;

#endif /* __WIN32_DW */

typedef unsigned short          DWC_tcp_port_t;
typedef struct sockaddr_in      DWC_tcp_socket_addr_t;

extern DWC_error_code_t     init_tcp_sockets ( void );
extern DWC_error_code_t     dinit_tcp_sockets ( void );
extern DWC_string_t         last_tcp_error ( void );
extern DWC_error_code_t     addr_tcp_op ( DWC_cmd_t cmd, DWC_tcp_socket_addr_t* addr, DWC_string_t name, va_list args );
extern DWC_error_code_t     create_tcp_socket ( DWC_tcp_socket_t *sock, DWC_bool_t nonblocking );
extern DWC_error_code_t     bind_tcp_socket ( DWC_tcp_socket_t sock, DWC_tcp_socket_addr_t* addr );
extern DWC_error_code_t     listen_tcp_socket ( DWC_tcp_socket_t sock, int backlog );
extern DWC_error_code_t     accept_tcp_socket ( DWC_tcp_socket_t sock, DWC_tcp_socket_addr_t* addr, DWC_tcp_socket_t *newsock );
extern DWC_error_code_t     connect_tcp_socket ( DWC_tcp_socket_t sock, DWC_tcp_socket_addr_t* addr, DWC_timeout_t timeout );
extern DWC_error_code_t     send_tcp_socket ( DWC_tcp_socket_t sock, DWC_pointer_t buf, DWC_length_t* len_ptr, int options );
extern DWC_error_code_t     recv_tcp_socket ( DWC_tcp_socket_t sock, DWC_pointer_t buf, DWC_length_t* len_ptr, int options );
extern DWC_error_code_t     wait_tcp_socket ( DWC_tcp_socket_t sock, int *how, DWC_timeout_t timeout );
extern DWC_error_code_t     shutdown_tcp_socket ( DWC_tcp_socket_t *sock );
extern DWC_error_code_t     close_tcp_socket ( DWC_tcp_socket_t *sock );
extern void                 cancel_tcp_call ( DWC_pointer_t sock );

#endif /* __DWC_TCP_H */
